from .config import *

qq_bot = None
already_started_discord_bot = False
discord_bot_threads = []
temp_bind_qq = {}  # qq:rid
temp_bind_discord = {}  # rid:dis
message_id_records = []  # (qq:dc)
forward_config = {}
loaded_forward_config = {}



from nonebot import logger
import httpx

from .. import global_vars as gv
from . import local as uLocal


async def webhook_send_message(username, avatar_url, content, fwd, images=[]):
    if images is None:
        images = []
    async with httpx.AsyncClient() as client:
        resp = await client.post(
            url=uLocal.get_discord_channel(fwd["discord-channel"])["webhook-url"] + "?wait=true",
            data={"username": username, "avatar_url": avatar_url, "content": content},
            files=[
                (
                    f"file[{idx+1}]",
                    (
                        f"file{idx+1}.jpg",
                        (await client.get(img_url)).content,
                        "image/jpeg",
                    ),
                )
                for idx, img_url in enumerate(images)
            ],
        )
        logger.debug("server response: " + str(resp.json()))
        return resp.json()["id"]


async def send_message_with_files(file_paths, name, content, fwd):
    async with httpx.AsyncClient() as client:
        files = []
        ix = 0
        for f in file_paths:
            files.append(
                (
                    f"file{ix}",
                    (f"image{ix}.jpg", (await client.get(url=f)).content, "image/jpeg"),
                )
            )
            ix += 1
        await client.post(
            url=f"https://discord.com/api/v10/channels/{uLocal.get_discord_channel(fwd['discord-channel'])['channel-id']}/messages",
            headers={"Authorization": f"Bot {uLocal.get_bot_token(uLocal.get_discord_channel(fwd['discord-channel'])['bot'])}"},
            data={"content": f"<{name}> {content}"},
            files=files,
        )


async def send_message(content, fwd):
    async with httpx.AsyncClient() as client:
        await client.post(
            url=f"https://discord.com/api/channels/{uLocal.get_discord_channel(fwd['discord-channel'])['channel-id']}/messages",
            headers={"Authorization": f"Bot {uLocal.get_bot_token(uLocal.get_discord_channel(fwd['discord-channel'])['bot'])}", "User-Agent": "DiscordBot"},
            data={"content": content},
        )


async def send_list_message(content_list, fwd):
    for i in content_list:
        if not i[1]:
            await send_message(i[0], fwd)
        else:
            await gv.qq_bot.send_group_msg(group_id=uLocal.get_qq_group_id(fwd['qq-group']), message=i[0])



import nonebot
from nonebot import logger
from nonebot.matcher import Matcher
from nonebot.adapters.onebot.v11 import Bot, GroupMessageEvent
from nonebot.adapters.onebot.v11.event import GroupRecallNoticeEvent
import threading
import httpx
import copy
import json
import os

from . import global_vars as gv
from .config import *
from .utils import forward as uForward
from .utils import local as uLocal
from .utils import send as uSend
from .utils import download as uDownload
from .utils import yamlloader as uYaml
from .bots import discordbot as bDiscord


os.environ["HTTP_PROXY"] = HTTP_PROXY
os.environ["HTTPS_PROXY"] = HTTP_PROXY
if not os.path.exists(qq_bind_file):
    json.dump({}, uLocal.safe_open(qq_bind_file, "w"))
if not os.path.exists(forwards_config_file):
    uYaml.dump(uYaml.default_config_data, forwards_config_file)
    logger.warning(f"转发配置文件已生成于 ( {os.path.abspath(forwards_config_file)} ) , 请配置后重启插件.")
    logger.warning("参考配置文件: https://github.com/xxtg666/nonebot-plugin-discord-message-bridge/blob/main/docs/dmb-config-example.yaml")
gv.forward_config = uYaml.load(forwards_config_file)
uLocal.load_forward_config()

if FORWARD_MSG_SERVER:
    enable_forward_msg_parse = True
else:
    enable_forward_msg_parse = False
    logger.warning(
        "未配置转发消息上传服务器, 将无法解析转发消息, 请在配置文件中设置 `dmb_forward_msg_server`"
    )


def set_qq_bind(discord_id, qq_id):
    qq_bind = json.load(open(qq_bind_file, "r"))
    qq_bind[str(discord_id)] = str(qq_id)
    json.dump(qq_bind, open(qq_bind_file, "w"))


@nonebot.on_message().handle()
async def _(matcher: Matcher, bot: Bot, event: GroupMessageEvent):
    try:
        group_id = event.group_id
    except Exception:
        return
    for fwd in uLocal.get_forwards(event.group_id, "qq-groups"):
        logger.debug(
            f"Received message from QQ: Message={str(event.get_message())} UserID={event.get_user_id()} GroupID={event.group_id} MessageID={event.message_id} "
            f"ReplyMessageID={event.reply.message_id if event.reply else None} SenderNickname={event.sender.nickname}"
        )
        if event.get_plaintext().startswith(QQ_COMMAND_PREFIX + QQ_COMMAND_NAME) and not fwd["silent"]:
            args = event.get_plaintext().strip().split(" ")[1:]
            if args:
                if args[0] == "bind":
                    try:
                        token = args[1]
                    except:
                        if dis_id := uLocal.get_qq_bind_discord(event.get_user_id()):
                            await matcher.finish(f"你已绑定 Discord({dis_id})", at_sender=True)
                        await matcher.finish(
                            f"你还未绑定 Discord, 请到消息转发频道下发送「{DISCORD_COMMAND_PREFIX}bind <uid>」进行绑定",
                            at_sender=True,
                        )
                    if (
                            token in gv.temp_bind_discord
                            and gv.temp_bind_qq[event.get_user_id()] == token
                    ):
                        set_qq_bind(gv.temp_bind_discord[token], event.get_user_id())
                        await uSend.send_message(
                            f"<@{gv.temp_bind_discord[token]}> QQ `{event.get_user_id()}` 绑定成功",
                            fwd
                        )
                        del gv.temp_bind_qq[event.get_user_id()]
                        del gv.temp_bind_discord[token]
                        await matcher.finish(f"Discord 绑定成功", at_sender=True)
                    else:
                        await matcher.finish("绑定 token 无效", at_sender=True)
                elif args[0] == "debug":
                    logger.info("Value of message_id_records: " + str(gv.message_id_records))
                    await matcher.finish("Success", at_sender=True)
                elif args[0] == "preview":
                    if event.reply:
                        reply = str(event.reply.message)
                        if reply.startswith("[CQ:forward") and enable_forward_msg_parse:
                            forward_msg_id = reply[15:-1].replace(",content=&", "")
                            chat_uuid = await uForward.get_forward_mapping(forward_msg_id)
                            await matcher.finish(uForward.get_preview_url(chat_uuid))

            await matcher.finish(
                BOT_NAME
                + " 命令帮助\n"
                + QQ_COMMAND_PREFIX + QQ_COMMAND_NAME
                + " bind <token> - 绑定 Discord 账户\n"
                + QQ_COMMAND_PREFIX + QQ_COMMAND_NAME
                + " debug - 在日志中获取 message_id_records",
                + QQ_COMMAND_PREFIX + QQ_COMMAND_NAME
                + " preview - 获取转发消息预览（回复转发消息时使用）",
                at_sender=True,
            )
        uid = event.get_user_id()
        message = event.get_message()
        origin_message = str(message)
        if origin_message.startswith("[CQ:forward") and enable_forward_msg_parse:
            forward_msg_id = origin_message[15:-1]
            forward_data = await uForward.get_forward_msg(forward_msg_id)
            chat_uuid = await uForward.upload_forward_msg(forward_data, forward_msg_id)
            preview_url = uForward.get_preview_url(chat_uuid)
            origin_message = f"[{FORWARD_MSG_PLACEHOLDER}]({preview_url})"
        else:
            origin_message = uLocal.process_text(origin_message)
        msg = uLocal.replace_cq_at_with_ids(origin_message)
        if msg.startswith(DISCORD_COMMAND_PREFIX * 4):
            await uSend.send_message(msg[2:], fwd)
            return
        msg_nocq = copy.deepcopy(msg)
        images = uLocal.get_url(msg)
        for i in uLocal.get_cq_images(msg):
            msg_nocq = msg_nocq.replace(i, IMAGE_PLACEHOLDER)
        if event.reply:
            if reply_to_dc_id := uLocal.get_another_message_id(
                event.reply.message_id, "qq"
            ):
                msg_content = (
                    (
                        await uDownload.get_discord_message_content(
                            reply_to_dc_id, fwd, removereply=True
                        )
                    )
                    .strip()
                    .replace("\n", " ")
                )
                msg_nocq = (
                    f"> {uLocal.generate_message_link(reply_to_dc_id, fwd)}\n> *{msg_content}*\n"
                    + msg_nocq
                )
        msg_id = await uSend.webhook_send_message(
            event.sender.nickname + SUFFIX, uLocal.get_qq_avatar_url(uid), msg_nocq, fwd, images
        )
        uLocal.record_message_id(event.message_id, msg_id)


@nonebot.on_notice().handle()
async def _(matcher: Matcher, bot: Bot, event: GroupRecallNoticeEvent):
    try:
        group_id = event.group_id
    except Exception:
        return
    for fwd in uLocal.get_forwards(event.group_id, "qq-groups"):
        if dc_id := uLocal.get_another_message_id(event.message_id, "qq"):
            async with httpx.AsyncClient() as client:
                await client.patch(
                    url=uLocal.get_discord_channel(fwd["discord-channel"])["webhook-url"] + "/messages/" + dc_id,
                    headers={"Content-Type": "application/json"},
                    json={
                        "content": "||"
                        + (await uDownload.get_discord_message_content(dc_id, fwd, e=False))
                        + "||"
                    },
                )


@nonebot.on_message().handle()
async def _(matcher: Matcher, bot: Bot):
    if not gv.already_started_discord_bot:
        gv.qq_bot = bot
        gv.already_started_discord_bot = True
        for bot_id, bot_token in gv.forward_config["discord-bots"].items():
            gv.discord_bot_threads.append(threading.Thread(target=bDiscord.startDiscordBot, args=(bot_token, bot_id)))
            gv.discord_bot_threads[-1].start()
            logger.success(f"Discord Bot {bot_id} thread started")
        matcher.destroy()



from ..config import *
from .. import global_vars as gv
import httpx


async def get_forward_msg(message_id):
    """使用nonebot2的方法获取合并转发消息"""
    try:
        # 使用nonebot2的call_api方法获取合并转发消息
        result = await gv.qq_bot.call_api("get_forward_msg", message_id=message_id)
        return {
            "data": result
        }
    except Exception as e:
        raise Exception(f"Failed to get forward message: {e}")

async def upload_forward_msg(data, id):
    async with httpx.AsyncClient() as client:
        response = await client.post(
            FORWARD_MSG_SERVER + "/upload/" + id,
            json=data
        )
        if response.status_code == 200:
            return response.json().get("chat_uuid", "")
        else:
            raise Exception("Failed to upload forward message")

def get_preview_url(uuid):
    return FORWARD_MSG_SERVER + "/?id=" + uuid

async def get_forward_mapping(id):
    async with httpx.AsyncClient() as client:
        response = await client.get(
            FORWARD_MSG_SERVER + "/mapping/" + id
        )
        if response.status_code == 200:
            return response.json().get("chat_uuid", "")
        else:
            raise Exception("Failed to get forward message mapping")


from nonebot.adapters.onebot.v11 import MessageSegment
from nonebot.adapters.onebot.v11.message import Message
from nonebot import logger
from discord import app_commands
import traceback
import discord
import asyncio

from .. import global_vars as gv
from ..config import *
from ..utils import send as uSend
from ..utils import local as uLocal
from ..utils import download as uDownload


intents = discord.Intents.default()
intents.message_content = True


def process_bind_command(discord_id, command, fwd):
    discord_id = str(discord_id)
    command = command.replace(DISCORD_COMMAND_PREFIX, "", 1)
    if command.startswith("bind"):
        try:
            qq_id = command.split(" ")[1]
            if qq_id in gv.temp_bind_qq:
                if gv.temp_bind_discord[gv.temp_bind_qq[qq_id]] == discord_id:
                    return [
                        (
                            f"<@{discord_id}> 你已在请求绑定QQ号,请在QQ群({uLocal.get_qq_group_id(fwd['qq-group'])})内发送 `{QQ_COMMAND_PREFIX+QQ_COMMAND_NAME} bind {gv.temp_bind_qq[qq_id]}` 进行绑定",
                            False,
                        ),
                        (
                            f"[CQ:at,qq={qq_id}] 请发送「{QQ_COMMAND_PREFIX+QQ_COMMAND_NAME} bind {gv.temp_bind_qq[qq_id]}」绑定 Discord({discord_id}) 如果这不是你的操作, 请忽略本条消息",
                            True,
                        ),
                    ]
                else:
                    del gv.temp_bind_qq[qq_id]
                    del gv.temp_bind_discord[gv.temp_bind_qq[qq_id]]
            rid = uLocal.genRandomID(8)
            gv.temp_bind_qq[qq_id] = rid
            gv.temp_bind_discord[rid] = discord_id
            return [
                (
                    f"<@{discord_id}> 请在QQ群({uLocal.get_qq_group_id(fwd['qq-group'])})内发送 `{QQ_COMMAND_PREFIX+QQ_COMMAND_NAME} bind {rid}` 进行绑定",
                    False,
                ),
                (
                    f"[CQ:at,qq={qq_id}] 请发送「{QQ_COMMAND_PREFIX+QQ_COMMAND_NAME} bind {rid}」绑定 Discord({discord_id}) 如果这不是你的操作, 请忽略本条消息",
                    True,
                ),
            ]
        except:
            if qq_id := uLocal.get_qq_bind(discord_id):
                return [(f"<@{discord_id}> 你已绑定QQ号 `{qq_id}`", False)]
            return [
                (
                    f"<@{discord_id}> 命令格式错误,应为: `{DISCORD_COMMAND_PREFIX}bind <uid>`",
                    False,
                )
            ]
    if not (qq_id := uLocal.get_qq_bind(discord_id)):
        return [
            (
                f"<@{discord_id}> 请先使用 `{DISCORD_COMMAND_PREFIX}bind <uid>` 绑定QQ号",
                False,
            )
        ]
    return [
        (
            f"<@{discord_id}> 您已绑定过QQ号 可以使用 `{DISCORD_COMMAND_PREFIX}bind <uid>` 修改绑定",
            False,
        )
    ]


def startDiscordBot(bot_token, bot_id):
    try:
        discord_client = discord.Client(intents=intents, proxy=HTTP_PROXY)

        @discord_client.event
        async def on_message(message):
            for fwd in uLocal.get_forwards(message.channel.id, "discord-channels"):
                if message.author.id == discord_client.user.id or str(
                    message.webhook_id
                ) == str(uLocal.get_discord_channel(fwd["discord-channel"])["webhook-id"]):
                    return
                logger.debug(
                    f"Received message from Discord: Message={message.content} UserID={message.author.id} ChannelID={message.channel.id} MessageID={message.id}"
                    f" ReplyMessageID={message.reference.message_id if message.reference else None}"
                )
                if message.content.startswith(DISCORD_COMMAND_PREFIX) and not fwd["silent"]:
                    await uSend.send_list_message(
                        process_bind_command(message.author.id, message.content, fwd), fwd
                    )
                    return
                ms = Message(
                    uLocal.replace_ids_with_cq_at(
                        fwd["discord-prefix"] + f"<{message.author.name}> {message.content}"
                    )
                )
                try:
                    async with (message.channel.typing() if not fwd["silent"] else uLocal.NoneAsyncWith()):
                        if message.attachments:
                            for atta in message.attachments:
                                ms += MessageSegment.image(
                                    await uDownload.download_image(atta.url)
                                )
                        if message.reference:
                            if reply_to_qq_id := uLocal.get_another_message_id(
                                message.reference.message_id, "dc"
                            ):
                                ms = MessageSegment.reply(int(reply_to_qq_id)) + ms
                        msg_id = (
                            await gv.qq_bot.send_group_msg(group_id=uLocal.get_qq_group_id(fwd['qq-group']), message=ms)
                        )["message_id"]
                        uLocal.record_message_id(msg_id, message.id)
                except:
                    if not fwd["silent"] and not NO_TRACEBACK:
                        await message.add_reaction(QQ_FORWARD_FAILED)
                        await message.reply("```" + traceback.format_exc() + "```")
                await asyncio.sleep(QQ_SEND_INTERVAL)

        @discord_client.event
        async def on_message_edit(before, after):
            for fwd in uLocal.get_forwards(before.channel.id, "discord-channels"):
                if before.content == after.content:
                    return
                if before.author.id == discord_client.user.id or str(
                    before.webhook_id
                ) == str(uLocal.get_discord_channel(fwd["discord-channel"])["webhook-id"]):
                    return
                async with (before.channel.typing() if not fwd["silent"] else uLocal.NoneAsyncWith()):
                    if qq_id := uLocal.get_another_message_id(before.id, "dc"):
                        ms = (
                            MessageSegment.reply(int(qq_id))
                            + EDIT_PLACEHOLDER
                            + after.content
                        )
                        msg_id = (
                            await gv.qq_bot.send_group_msg(group_id=uLocal.get_qq_group_id(fwd['qq-group']), message=ms)
                        )["message_id"]
                        uLocal.record_message_id(msg_id, before.id)
                await asyncio.sleep(QQ_SEND_INTERVAL)

        @discord_client.event
        async def on_message_delete(message):
            for fwd in uLocal.get_forwards(message.channel.id, "discord-channels"):
                if message.author.id == discord_client.user.id or str(
                    message.webhook_id
                ) == str(uLocal.get_discord_channel(fwd["discord-channel"])["webhook-id"]):
                    return
                async with (message.channel.typing() if not fwd["silent"] else uLocal.NoneAsyncWith()):
                    if qq_id := uLocal.get_another_message_id(message.id, "dc"):
                        ms = MessageSegment.reply(int(qq_id)) + DELETE_PLACEHOLDER
                        msg_id = (
                            await gv.qq_bot.send_group_msg(group_id=uLocal.get_qq_group_id(fwd['qq-group']), message=ms)
                        )["message_id"]
                        uLocal.record_message_id(msg_id, message.id)
                await asyncio.sleep(QQ_SEND_INTERVAL)

        discord_tree = app_commands.CommandTree(discord_client)

        @discord_tree.command(name="debug-dmb", description="Log message_id_records & loaded_forward_config")
        async def debug_dmb(interaction):
            logger.info("Value of message_id_records: " + str(gv.message_id_records))
            logger.info("Value of loaded_forward_config: " + str(gv.loaded_forward_config))
            await interaction.response.send_message(f"Success")

        @discord_client.event
        async def on_ready():
            await discord_tree.sync()
            logger.success(f"Discord Bot {bot_id} started")

        discord_client.run(bot_token)
    except:
        logger.error(traceback.format_exc())



from pydantic import BaseModel
from nonebot import get_plugin_config


class Config(BaseModel):
    # 网络代理地址，若不需要请留空
    dmb_http_proxy: str = "http://127.0.0.1:7890"
    
    # QQ - Discord 用户 id 绑定文件路径
    dmb_qq_bind_file: str = "data/discord_message_bridge_qq_bind.json"
    
    # 转发配置文件路径
    dmb_forwards_config_file: str = "data/discord_message_bridge_forwards.yaml"
    
    # 记录 QQ 与 Discord 消息互相对应 id 的最大数量
    dmb_max_message_id_record: int = 6000
    
    # 转发失败时添加的 reaction 名称 (支持自定义表情符号)
    dmb_forward_failed_reaction: str = "😢"
    
    # 转发回复消息时显示的原消息内容的最大长度
    dmb_max_reply_preview_length: int = 100
    
    # Discord -> QQ 转发消息前缀
    dmb_prefix: str = "[Discord] "
    
    # QQ -> Discord 转发消息后缀
    dmb_suffix: str = " [QQ]"
    
    # 转发图片占位符
    dmb_image_placeholder: str = " [图片] "
    
    # 编辑消息占位符
    dmb_edit_placeholder: str = "[编辑消息] "
    
    # 消息已被删除占位符
    dmb_delete_placeholder: str = "[消息已被删除]"
    
    # Bot 名称
    dmb_bot_name: str = "Discord Message Bridge"
    
    # QQ 机器人命令名
    dmb_qq_command: str = "dmb"
    
    # Discord 机器人命令前缀
    dmb_discord_command_prefix: str = "~"
    
    # QQ 机器人命令前缀
    dmb_qq_command_prefix: str = "."
    
    # 关闭报错反馈
    dmb_no_traceback: bool = False
    
    # 一对多发送消息间隔
    dmb_qq_send_interval: float = 0.2
    
    # 合并转发消息回复占位符
    dmb_forward_msg_placeholder: str = "[合并转发]"
    
    # 转发消息上传服务器，请部署 https://github.com/xxtg666/Forward-Message-Server 后修改为其 SERVER_URL
    dmb_forward_msg_server: str = ""


config = get_plugin_config(Config)

HTTP_PROXY = config.dmb_http_proxy
QQ_FORWARD_FAILED = config.dmb_forward_failed_reaction
qq_bind_file = config.dmb_qq_bind_file
forwards_config_file = config.dmb_forwards_config_file
MAX_REPLY_PREVIEW_LENGTH = config.dmb_max_reply_preview_length
MAX_MESSAGE_ID_RECORD = config.dmb_max_message_id_record
PREFIX = config.dmb_prefix
SUFFIX = config.dmb_suffix
IMAGE_PLACEHOLDER = config.dmb_image_placeholder
EDIT_PLACEHOLDER = config.dmb_edit_placeholder
DELETE_PLACEHOLDER = config.dmb_delete_placeholder
BOT_NAME = config.dmb_bot_name
QQ_COMMAND_NAME = config.dmb_qq_command
QQ_COMMAND_PREFIX = config.dmb_qq_command_prefix
DISCORD_COMMAND_PREFIX = config.dmb_discord_command_prefix
NO_TRACEBACK = config.dmb_no_traceback
QQ_SEND_INTERVAL = config.dmb_qq_send_interval
FORWARD_MSG_PLACEHOLDER = config.dmb_forward_msg_placeholder
FORWARD_MSG_SERVER = config.dmb_forward_msg_server
